%%%% Data centre system model, as described in Tatchell-Evans (2017, in press)
%%%% Free cooling version (no mechanical cooling)
clear all

CRACRefMultiplier=1; % Set to 2 to double size of CRAC compared to default
T_amb=284; % Mean temperature for London=11C
T_PA_Sup=298; % Process air supply temperature (K)
P_CA=7; % cold aisle pressure (pa)
N_servers=308; % number of servers (308 gives 12kW/rack using Barroso et al's (2007) CPU utilisation data)
C_D=38289; % discharge coefficient of server, from Brady (2016, p203) (Pa/m^3/s)
FansPServer=8; % Number of server fans per server
ServerFanPowerMax=0.006; % Max power (W) of a single server fan
ITPower=(0.23617*N_servers/10)*[0.51 0.548 0.595 0.625 0.633 0.66 0.683 0.71 0.745 0.883]; % From Brady (2016), v20z consumes 236.17 W at 100% utilisation with dp=0.  Distribution of server power consumptions as per data from Barroso et al. 2007.
ITPower_0=0.23617*N_servers; % Assume infrastructure is sized to cope with all servers operating at 100%
dT_IT=12.5; % temperature rise across servers
c_pAir=1.005; % specific heat capacity of air (kJ/kgK)
c_pWater=4.18; % specific heat capacity of water (kJ/kgK)
rho_Water=1000; % density of water (kg/m^3)
rho_Air=1.2; % density of air (kg/m^3)
QServers_0=ITPower_0/(rho_Air*c_pAir*dT_IT); % IT flow rate required with all servers operating at 100%, dp=0.
eta_CRACFan=0.9; % Efficiency of CRAC fans, consistent with Salim & Tozer (2010)
eta_CWPump=0.9; % Efficiency of chilled water pumps, consistent with Salim & Tozer (2010)
CWPumpPowerRef=2*60.91*0.0014/eta_CWPump; % kW. Uniflair TD/UCV 0700 has a pressure drop of 60.91 kPa at 1.4 l/s.  Assume same pressure drop in the economiser.
CWPumpPowerRef=CWPumpPowerRef*ITPower_0*1.1/31.2; % Assume reference CW pump power scales linearly from Uniflair data for condition where load on chilled water loop is 1.1x IT load.  Will then scale from there by the cube of flow rate.  CW unit provides 31.2 kW of cooling at reference flow rate
Q_CWRef=0.0014*ITPower_0*1.1/31.2; % referenced Uniflair data corresponds to CW flow rate of 1.4 l/s, providing 31.2kW of cooling.  Assume heat exchangers etc. are scaled up appropriately, to cool the housed IT at a PUE of 1.1.
Q_PARef=1.66; % referenced Uniflair data corresponds to air flow rate of 1.66 m^3/s
h_CRACRef=23.9/7.27; % The Uniflair TD/UCV achieves a sensible heat transfer of 23.9kW at LMTD=7.27
Force=0; % To allow server fan speeds to reduce with cold aisle pressure set this to 0.  To ignore cold aisle pressure when calculating fan speed, set to 1.  Hence '0' corresponds to Server Fan Option 2, '1' corresponds to Server Fan Option 1.
index=0.631; % index for relationship between h_CRAC and Reynolds number (figures from the literature range from 0.581-0.681)

% open server fan curve data (data from Brady, 2016)
Fan25=transpose([0.004 0.004 0.0033 0.0025 0.002 0.002 0 0; 0 6.3 8.3 13.5 20.8 27.1 33 39.6]);
Fan50=transpose([0.0049 0.0049 0.0049 0.0045 0.0034 0.0025 0.0025 0.0025 0.002 0 0; 0 2.1 5.2 8.3 18.8 29 31.3 35.4 42.7 54.2 63.5]);
Fan75=transpose([0.0079 0.0079 0.0071 0.0064 0.0049 0.0045 0.004 0.004 0.0025 0.002 0; 0 6.3 14.6 25 44.8 64.6 69.8 80.2 94.8 125 144.8]);
Fan100=transpose([0 0.0092 0.0092 0.0087 0.0061 0.0053 0.0049 0.004 0.0034 0.0028 0; 0 7.3 9.4 21.9 67.7 95.8 105 120 141 191 235]);

Q_serversRequired=ITPower*(1/(dT_IT*1.005*1.2)); % m^3/s assuming delta T is 12.5, heat capacity is 1.005 and density is 1.2
Q_serversForced=N_servers*sqrt(P_CA/C_D); % m^3/s flow rate through server due to cold aisle pressure

if Force==0 % if flow forced through the server by cold aisle pressure is assumed to lead to reduced server fan speeds
iteration=0;
    
    for j=1:length(ITPower)
        if Q_serversForced/length(ITPower)>=Q_serversRequired(j) % i.e. if air forced through server/server block is sufficient
            Q_servers(j)=Q_serversForced/length(ITPower); % assume server fans switch off
            SFanUtil=0;
            P_required=0;
            ServerFanPower=0;
            
            % Now calculate what server fan power consumption would be
            % without containment
            Q_F0=zeros(4,1); % This to be populated with the flow rates achieved by the fan at this pressure, at 25, 50, 75 and 100% of server fan power
                for i=1:length(Fan25)
                    if Fan25(i,2)>P_required
                        Q_F0(1)=Fan25(i-1,1)+(Fan25(i,1)-Fan25(i-1,1))*(P_required-Fan25(i-1,2))/(Fan25(i,2)-Fan25(i-1,2));
                        break
                    end
                end
                for i=1:length(Fan50)
                    if Fan50(i,2)>P_required
                        Q_F0(2)=Fan50(i-1,1)+(Fan50(i,1)-Fan50(i-1,1))*(P_required-Fan50(i-1,2))/(Fan50(i,2)-Fan50(i-1,2));
                        break
                    end
                end
                for i=1:length(Fan75)
                    if Fan75(i,2)>P_required
                        Q_F0(3)=Fan75(i-1,1)+(Fan75(i,1)-Fan75(i-1,1))*(P_required-Fan75(i-1,2))/(Fan75(i,2)-Fan75(i-1,2));
                        break
                    end
                end
                for i=1:length(Fan100)
                    if Fan100(i,2)>P_required
                        Q_F0(4)=Fan100(i-1,1)+(Fan100(i,1)-Fan100(i-1,1))*(P_required-Fan100(i-1,2))/(Fan100(i,2)-Fan100(i-1,2));
                        break
                    end
                end

                for i=1:length(Q_F0)
                    % Repeat for fan power consumption without containment
                    if FansPServer*(N_servers/length(ITPower))*Q_F0(i)>Q_serversRequired(j) % if flow achieved at this (25%, 50%, 75% or 100%) server fan power consumption is sufficient
                        if i==1
                            ServerFanPower0=ServerFanPowerMax*FansPServer*(N_servers/length(ITPower))*((Q_serversRequired(j)/(FansPServer*(N_servers/length(ITPower))))/(Q_F0(i)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
                        else
                            ServerFanPower0=ServerFanPowerMax*FansPServer*(N_servers/length(ITPower))*(i-1+((Q_serversRequired(j)/(FansPServer*(N_servers/length(ITPower))))-Q_F0(i-1))/(Q_F0(i)-Q_F0(i-1)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
                        end
                        break
                    end
                end
                SFanUtil(j)=0; % Proportion of server fan power in use
                Q_servers(j)=(N_servers/length(ITPower))*sqrt((P_CA)/C_D);
                ServerFanEta(j)=0;
                ServerFanHeat(j)=0; % Heat created due to inefficiency of fans, eta=(flow energy)/(fan power)
        else % i.e. if server fans are required
            P_required=C_D*(Q_serversRequired(j)/(N_servers/length(ITPower)))^2; % (Pa) pressure required for sufficient server air flow
            P_excess=P_required-P_CA; % pressure generation required of server fans
            % calculate server fan power required
            Q_F=zeros(4,1); % This to be populated with the flow rates achieved by the fan at this pressure, at 25, 50, 75 and 100% of server fan power
                for i=1:length(Fan25)
                    if Fan25(i,2)>P_excess
                        Q_F(1)=Fan25(i-1,1)+(Fan25(i,1)-Fan25(i-1,1))*(P_excess-Fan25(i-1,2))/(Fan25(i,2)-Fan25(i-1,2));
                        break
                    end
                end
                for i=1:length(Fan50)
                    if Fan50(i,2)>P_excess
                        Q_F(2)=Fan50(i-1,1)+(Fan50(i,1)-Fan50(i-1,1))*(P_excess-Fan50(i-1,2))/(Fan50(i,2)-Fan50(i-1,2));
                        break
                    end
                end
                for i=1:length(Fan75)
                    if Fan75(i,2)>P_excess
                        Q_F(3)=Fan75(i-1,1)+(Fan75(i,1)-Fan75(i-1,1))*(P_excess-Fan75(i-1,2))/(Fan75(i,2)-Fan75(i-1,2));
                        break
                    end
                end
                for i=1:length(Fan100)
                    if Fan100(i,2)>P_excess
                        Q_F(4)=Fan100(i-1,1)+(Fan100(i,1)-Fan100(i-1,1))*(P_excess-Fan100(i-1,2))/(Fan100(i,2)-Fan100(i-1,2));
                        break
                    end
                end
                
                % Here the process is repeated but looking for the server
                % fan power consumption which would be required if
                % containment were not in place
                Q_F0=zeros(4,1); % This to be populated with the flow rates achieved by the fan at this pressure, at 25, 50, 75 and 100% of server fan power
                for i=1:length(Fan25)
                    if Fan25(i,2)>P_required
                        Q_F0(1)=Fan25(i-1,1)+(Fan25(i,1)-Fan25(i-1,1))*(P_required-Fan25(i-1,2))/(Fan25(i,2)-Fan25(i-1,2));
                        break
                    end
                end
                for i=1:length(Fan50)
                    if Fan50(i,2)>P_required
                        Q_F0(2)=Fan50(i-1,1)+(Fan50(i,1)-Fan50(i-1,1))*(P_required-Fan50(i-1,2))/(Fan50(i,2)-Fan50(i-1,2));
                        break
                    end
                end
                for i=1:length(Fan75)
                    if Fan75(i,2)>P_required
                        Q_F0(3)=Fan75(i-1,1)+(Fan75(i,1)-Fan75(i-1,1))*(P_required-Fan75(i-1,2))/(Fan75(i,2)-Fan75(i-1,2));
                        break
                    end
                end
                for i=1:length(Fan100)
                    if Fan100(i,2)>P_required
                        Q_F0(4)=Fan100(i-1,1)+(Fan100(i,1)-Fan100(i-1,1))*(P_required-Fan100(i-1,2))/(Fan100(i,2)-Fan100(i-1,2));
                        break
                    end
                end

                for i=1:length(Q_F)
                    if FansPServer*(N_servers/length(ITPower))*Q_F(i)>Q_serversRequired(j) % if flow achieved at this (25%, 50%, 75% or 100%) server fan power consumption is sufficient
                        if i==1
                            ServerFanPower=ServerFanPowerMax*FansPServer*(N_servers/length(ITPower))*((Q_serversRequired(j)/(FansPServer*(N_servers/length(ITPower))))/(Q_F(i)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
                        else
                            ServerFanPower=ServerFanPowerMax*FansPServer*(N_servers/length(ITPower))*(i-1+((Q_serversRequired(j)/(FansPServer*(N_servers/length(ITPower))))-Q_F(i-1))/(Q_F(i)-Q_F(i-1)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
                        end
                        break
                    end

                end
                % Repeat for fan power consumption without containment
                for i=1:length(Q_F0)
                    if FansPServer*(N_servers/length(ITPower))*Q_F0(i)>Q_serversRequired(j) % if flow achieved at this (25%, 50%, 75% or 100%) server fan power consumption is sufficient
                        if i==1
                            ServerFanPower0=ServerFanPowerMax*FansPServer*(N_servers/length(ITPower))*((Q_serversRequired(j)/(FansPServer*(N_servers/length(ITPower))))/(Q_F0(i)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
                        else
                            ServerFanPower0=ServerFanPowerMax*FansPServer*(N_servers/length(ITPower))*(i-1+((Q_serversRequired(j)/(FansPServer*(N_servers/length(ITPower))))-Q_F0(i-1))/(Q_F0(i)-Q_F0(i-1)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
                        end
                        break
                    end
                end
        SFanUtil(j)=(ServerFanPower/(FansPServer*(N_servers/length(ITPower))))/ServerFanPowerMax; % Proportion of server fan power in use
        Q_servers(j)=(N_servers/length(ITPower))*sqrt((P_excess+P_CA)/C_D);
        ServerFanEta(j)=Q_serversRequired(j)*P_required/(1000*ServerFanPower);
        ServerFanHeat(j)=ServerFanPower*(1-ServerFanEta(j)); % Heat created due to inefficiency of fans, eta=(flow energy)/(fan power)
        end
        ITPower(j)=ITPower(j)-(ServerFanPower0-ServerFanPower); % Account for the reduction in server power consumption as a result of cold aisle pressure's effect on server fan speed
    end
    Q_serversTotal=sum(Q_servers); % sum the flow through each block of servers
    
    
elseif Force==1 % if server fan speed is assumed to depend only on CPU utilisation, not cold aisle pressure
    for j=1:length(ITPower)
        P_required=C_D*(Q_serversRequired(j)/(N_servers/length(ITPower)))^2; % (Pa) pressure required for sufficient server air flow
        Q_servers(j)=(N_servers/length(ITPower))*sqrt((P_required+P_CA)/C_D); % m^3/s flow rate through server due to sum of cold aisle pressure & pressure induced by server fans

        % Now calculate server fan power
        Q_F=zeros(4,1); % This to be populated with the flow rates achieved by the fan at this pressure, at 25, 50, 75 and 100% of server fan power
        for i=1:length(Fan25)
            if Fan25(i,2)>P_required
                Q_F(1)=Fan25(i-1,1)+(Fan25(i,1)-Fan25(i-1,1))*(P_required-Fan25(i-1,2))/(Fan25(i,2)-Fan25(i-1,2));
                break
            end
        end
        for i=1:length(Fan50)
            if Fan50(i,2)>P_required
                Q_F(2)=Fan50(i-1,1)+(Fan50(i,1)-Fan50(i-1,1))*(P_required-Fan50(i-1,2))/(Fan50(i,2)-Fan50(i-1,2));
                break
            end
        end
        for i=1:length(Fan75)
            if Fan75(i,2)>P_required
                Q_F(3)=Fan75(i-1,1)+(Fan75(i,1)-Fan75(i-1,1))*(P_required-Fan75(i-1,2))/(Fan75(i,2)-Fan75(i-1,2));
                break
            end
        end
        for i=1:length(Fan100)
            if Fan100(i,2)>P_required
                Q_F(4)=Fan100(i-1,1)+(Fan100(i,1)-Fan100(i-1,1))*(P_required-Fan100(i-1,2))/(Fan100(i,2)-Fan100(i-1,2));
                break
            end
        end

        for i=1:length(Q_F)
            if FansPServer*(N_servers/length(ITPower))*Q_F(i)>Q_serversRequired(j)  % if flow achieved at this (25%, 50%, 75% or 100%) server fan power consumption is sufficient
                if i==1
                    ServerFanPower=ServerFanPowerMax*FansPServer*(N_servers/length(ITPower))*((Q_serversRequired(j)/(FansPServer*(N_servers/length(ITPower))))/(Q_F(i)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
                else
                    ServerFanPower=ServerFanPowerMax*FansPServer*(N_servers/length(ITPower))*(i-1+((Q_serversRequired(j)/(FansPServer*(N_servers/length(ITPower))))-Q_F(i-1))/(Q_F(i)-Q_F(i-1)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
                end
                break
            end
        end
        SFanUtil(j)=(ServerFanPower/(FansPServer*(N_servers/length(ITPower))))/ServerFanPowerMax; % Proportion of server fan power in use
        ServerFanEta(j)=Q_serversRequired(j)*P_required/(1000*ServerFanPower);
        ServerFanHeat(j)=ServerFanPower*(1-ServerFanEta(j)); % Heat created due to inefficiency of fans, eta=(flow energy)/(fan power)
    end
    Q_serversTotal=sum(Q_servers);
end

Q_serversExcess=Q_serversTotal-sum(Q_serversRequired);

Bypass=(((-80.038+(80.038^2+4*452.89*P_CA)^0.5)/(2*452.89))+0.75*((-56.592+(56.592^2+4*1289*P_CA)^0.5)/(2*1289))); % Bypass per rack, for low bypass condition with server interface leakage.  Tatchell-Evans 2017, chapter 2.
%Bypass=((-3.689+(3.689^2+4*45.855*P_CA)^0.5)/(2*45.855)); % Bypass per rack, for high bypass condition.  Tatchell-Evans 2017, chapter 2.
Bypass=Bypass*4; % Scaled up to 4 rack system, as in Test Data Centre

T_PA_Ret=T_PA_Sup+sum(ITPower)/((Q_serversTotal+Bypass)*rho_Air*c_pAir); % Temperature (K) of process air returning to computer room air conditioner (CRAC)
Q_CRACRef=QServers_0*1.2*CRACRefMultiplier; % For calculating CRAC fan power consumption, assuming CRAC is sized for all servers operating at 100% utilisation, plus 20% bypass.
CRACFanPower=(Q_CRACRef/1.66)*2.2*(((Q_serversTotal+Bypass)/Q_CRACRef)^3); % Power consumption of CRAC fans (kW).  Uniflair TD/UCV 0700 fan uses 2.2kW at 1.66m^3/s.  Scale this with the reference CRAC flow rate to get reference CRAC fan power, then scale with cube of flow rate for variation.
TotalServerFanPower=mean(SFanUtil)*N_servers*FansPServer*ServerFanPowerMax;
CRACLoad=sum(ITPower)-TotalServerFanPower+sum(ServerFanHeat);

% heat exchanger calculations
h_CRACRef=h_CRACRef*Q_CRACRef/1.66; % Assume heat transfer per LMTD scales linearly with size of CRAC
h_CRAC=h_CRACRef*((Q_serversTotal+Bypass)/Q_CRACRef)^index;
% Now test over a range of dT_CW
i=0;
V_CWControl=0; % This variable simply records whether or not CW flow rate was controlled to minimise power consumption
Q_EcoRef=Q_CRACRef*2;

for dT_CW=1:10 % for CW temperature rise of 1 to 10
    i=i+1;
    Q_CW(i)=(CRACLoad+CRACFanPower*(1-eta_CRACFan))/(dT_CW*c_pWater*rho_Water); % calculate minimum CW flow rate required to achieve the heat transfer at this CW dT
    CWPumpPower(i)=CWPumpPowerRef*(Q_CW(i)/Q_CWRef)^3; % calculate CW pump power assuming it scales with cube of flow rate
    
    LMTD_required(i)=(CRACLoad+CRACFanPower*(1-eta_CRACFan))/h_CRAC; % This LMTD is required to achieve the heat transfer on the air side
    % Finding CW supply temperature requires numerical methods:
    %T_CWh=T_CWc+dT_CW;
    %dTa=T_PA_Ret-T_CWh;    
    digitsOld=digits;
    digits(6)
    syms T_CWc
    T_CWc=double(vpasolve((T_PA_Ret-(T_CWc+dT_CW)-(T_PA_Sup-T_CWc))/log((T_PA_Ret-(T_CWc+dT_CW))/(T_PA_Sup-T_CWc)) == LMTD_required(i), T_CWc));
    digits(digitsOld)
    T_CWcStore(i)=T_CWc;
    
    EcoLoad(i)=CRACLoad+CRACFanPower*(1-eta_CRACFan)+CWPumpPower(i)*(1-eta_CWPump); % Load on the economiser
    dTb=T_CWc-T_amb;
    % Noting that h_eco=h_CRACRef * ((Q_CW/Q_CWRef)^0.8) * (((Q_Eco)/Q_EcoRef)^index)
    % and T_EA_Ex=T_amb+EcoLoad/(rho_Air*Q_Eco*c_pAir)
    % also EcoLoad=h_eco(LMTD_eco)
    % LMTD_eco=(((T_CWc+dT_CW)-(T_amb+EcoLoad/(rho_Air*Q_EA*c_pAir)))-dTb)/log(((T_CWc+dT_CW)-(T_amb+EcoLoad/(rho_Air*Q_Eco*c_pAir)))/dTb)
    digitsOld=digits;
    digits(6)
    syms Q_Eco
    % Combining the above equations allows us to solve numerically for the
    % flow rate required of the economiser fans:
    Q_Eco=double(vpasolve((h_CRACRef*((Q_CW(i)/Q_CWRef)^0.8)*(((Q_Eco)/Q_EcoRef)^index)) * (((T_CWc+dT_CW)-(T_amb+EcoLoad(i)/(rho_Air*Q_Eco*c_pAir)))-dTb)/log(((T_CWc+dT_CW)-(T_amb+EcoLoad(i)/(rho_Air*Q_Eco*c_pAir)))/dTb) == EcoLoad(i), T_CWc));
    digits(digitsOld)
    
    Q_EcoStore(i)=Q_Eco;
    EcoFanPower(i)=(Q_EcoRef/1.66)*2.2*((Q_Eco/Q_EcoRef)^3); % Power consumption of economiser fans (kW).  Uniflair TD/UCV 0700 fan uses 2.2kW at 1.66m^3/s.  Scale this with the reference CRAC flow rate to get reference fan power, then scale with cube of flow rate for variation.
    Power(i)=[EcoFanPower(i)+CWPumpPower(i)];
    V_CWControl=1; % Setting this variable to 1 indicates that CW flow rate was controlled to minimise power consumption
end

[MinPower i]=min(Power); % find index for combined power consumption of CW pump and economiser fan

TotalPower=MinPower+CRACFanPower+sum(ITPower);

Results=vertcat([V_CWControl mean(SFanUtil) ServerFanPower sum(ITPower) Q_serversTotal Q_serversExcess Bypass CRACFanPower CWPumpPower(i) TotalPower]);
COP_C=CRACLoad/(TotalPower-sum(ITPower)-CRACFanPower)