 % PF-CFD model of example data centre, as described in Chapter 5 of
 % Tatchell-Evans 2017 (in press)

clear all
Counter=0;

TimeStart=[datestr(now, 'HH') datestr(now, 'MM')]

PhiIt=2;

% Fix supply flow rate or p_CA?
prompt='Fix supply flow rate or cold aisle pressurisation? 1 for flow rate, 2 for pressure.';
EquilCond=input(prompt);
if EquilCond==1
    prompt='Set supply flow rate in m^3/s';
    VIn=input(prompt);
else
    prompt='Set pressure drop from cold to hot aisle in Pa';
    p_CA=input(prompt);
end


%% Properties of air etc., bypass and server flows, supply air conditions
rho_0=1.2; % kg/m^3
c_p=1.005; % kJ/kg.K
k_h=0.0264*10^-3; % Heat transfer coefficient, kW/mK
p_0=101000; % Atmospheric pressure, Pa
T_0=293; % Supply temperature, K

%% Dimensions etc.
% Original cell numbers
N_racks=16; % Number of racks in aisle
ServPerRack=21; % Number of servers in each rack
% 0.1m sided cells
CellSide=0.1;
nx=round(N_racks*0.8/CellSide);
ny=round(3/CellSide); 
nz=round(2.4/CellSide); 
L=nx*CellSide;
H=ny*CellSide;
D=nz*CellSide;

Dx=L/nx; % mesh increment in the x direction (m)
Dy=H/ny; % mesh increment in the y direction (m)
Dz=D/nz; % mesh increment in the z direction (m)

%%%%% For test data centre inlet
% AInlet=pi*(0.315/2)^2; % Area of inlet
% HInlet=AInlet^0.5; % Height of inlet
% DInlet=AInlet^0.5; % Width of inlet
% Inlety=0.845; %y-offset of inlet
% Inletz=0.665; % z-offset of inlet

%%%%% Inlet matching rack height and full width of cold aisle
HInlet=3;
DInlet=0.6;
WInlet=0;
Outletx=0;
Inlety=Dy/2;
Inletz=Dz/2;
InWall=1;

%%%%%%% Outlet matching rack height and full width of hot aisle
HOutlet=3; % Height of outlet
DOutlet=0.6; % Width of outlet
WOutlet=0; % Depth of outlet
Outletx=0; % x-offset of outlet
Outlety=0; %y-offset of outlet
Outletz=1.8; % z-offset of outlet
OutWall=2; % Outlet is on wall opposite supply

Contz=0.7; % z-position of fronts of servers - 0.1m back from where rack front would be
Servx=((0:N_racks-1)*0.8+0.2); % Array containing x offsets of racks 1 to N_racks
Servy=0.1*ones(1,N_racks); % Array containing y offsets of racks 1 to N_racks
HServ=ServPerRack*0.04445*ones(1,N_racks); % Array containing heights of servers in racks 1 to N_racks (1.8669=42U)
WServ=0.4*ones(1,N_racks); % width of server (m)
LServ=0.8*ones(1,N_racks); % length of server (m)
EServ_0=0.23617*ServPerRack*ones(1,N_racks); % Just for initialisation. From Gemma Brady's thesis (2016, p217), 1U v20z server uses ~236.17W at 0 ESP, when stressed.
VServ_0=(EServ_0/(c_p*12.5)).*ones(1,N_racks); % Just for initialisation, conservation of heat assuming dT=12.5, and assuming server fan speed consistent with stressed condition at dp=0
RackDepth=1.1; % Depth of rack
RackHeight=2.1; % Height of rack
ServerPts(N_racks)=0;
EmptySlot=0; % Empty slot above server? 1=yes.
SlotPts(N_racks)=0;
BypassPts=0; % Number of nodes through which bypass is allowed

if EquilCond==1
    p_CA=(VIn>sum(VServ_0))*2-(VIn<sum(VServ_0))*2; % pressure initially defaults to 2 or -2, depending on whether supply flow exceeds total IT flow or is less than it
else
    VIn=(1+p_CA/50)*sum(VServ_0); % Initialise supply flow rate
end
VInStore=VIn;
p_CA_store=p_CA;

%Now calculate server fan power consumption at local dp=0
C_D=38289; % discharge coefficient of v20z server, from Brady (2016, p203) (Pa/m^3/s)
P_required=C_D*(VServ_0/ServPerRack).^2; % Pressure to be developed by internal server fans
% open server fan curve data (from Brady (2016))
Fan25=transpose([0.004 0.004 0.0033 0.0025 0.002 0.002 0 0; 0 6.3 8.3 13.5 20.8 27.1 33 39.6]);
Fan50=transpose([0.0049 0.0049 0.0049 0.0045 0.0034 0.0025 0.0025 0.0025 0.002 0 0; 0 2.1 5.2 8.3 18.8 29 31.3 35.4 42.7 54.2 63.5]);
Fan75=transpose([0.0079 0.0079 0.0071 0.0064 0.0049 0.0045 0.004 0.004 0.0025 0.002 0; 0 6.3 14.6 25 44.8 64.6 69.8 80.2 94.8 125 144.8]);
Fan100=transpose([0 0.0092 0.0092 0.0087 0.0061 0.0053 0.0049 0.004 0.0034 0.0028 0; 0 7.3 9.4 21.9 67.7 95.8 105 120 141 191 235]);

% Calculate server fan power required at 0 external static pressure
Q_F=zeros(4,1); % This to be populated with the flow rates achieved by the fan at this pressure, at 25, 50, 75 and 100% of server fan power
for j=1:length(P_required)
    for i=1:length(Fan25)
        if Fan25(i,2)>P_required(j)
            Q_F(1)=Fan25(i-1,1)+(Fan25(i,1)-Fan25(i-1,1))*(P_required(j)-Fan25(i-1,2))/(Fan25(i,2)-Fan25(i-1,2));
            break
        end
    end
    for i=1:length(Fan50)
        if Fan50(i,2)>P_required(j)
            Q_F(2)=Fan50(i-1,1)+(Fan50(i,1)-Fan50(i-1,1))*(P_required(j)-Fan50(i-1,2))/(Fan50(i,2)-Fan50(i-1,2));
            break
        end
    end
    for i=1:length(Fan75)
        if Fan75(i,2)>P_required(j)
            Q_F(3)=Fan75(i-1,1)+(Fan75(i,1)-Fan75(i-1,1))*(P_required(j)-Fan75(i-1,2))/(Fan75(i,2)-Fan75(i-1,2));
            break
        end
    end
    for i=1:length(Fan100)
        if Fan100(i,2)>P_required(j)
            Q_F(4)=Fan100(i-1,1)+(Fan100(i,1)-Fan100(i-1,1))*(P_required(j)-Fan100(i-1,2))/(Fan100(i,2)-Fan100(i-1,2));
            break
        end
    end
    FansPServer=8; % This is the number of fans in each v20z server, from Brady
    ServerFanPowerMax=0.006; % Power consumption of a single server fan, from Brady
    for i=1:length(Q_F)
        if FansPServer*ServPerRack*Q_F(i)>VServ_0(j) % if flow achieved at this (25%, 50%, 75% or 100%) server fan power consumption is sufficient
            if i==1
                ServerFanPower(j)=ServerFanPowerMax*FansPServer*ServPerRack*((VServ_0(j)/(FansPServer*ServPerRack))/(Q_F(i)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
            else
                ServerFanPower(j)=ServerFanPowerMax*FansPServer*ServPerRack*(i-1+((VServ_0(j)/(FansPServer*ServPerRack))-Q_F(i-1))/(Q_F(i)-Q_F(i-1)))/4; % server fan power in Watts (divided by 4 because i=1 corresponds to 25% power consumption)
            end
            break
        end

    end
end


if HInlet<2*Dy || DInlet<2*Dz
    'ERROR - Inlet too small in comparison to grid spacing'
elseif HOutlet<2*Dy || DOutlet<2*Dz
    'ERROR - Outlet too small in comparison to grid spacing'
elseif Inlety<Dy || Inletz<Dz || Inlety+HInlet>=H || Inletz+DInlet>=D
    'ERROR - Inlet does not fit within domain'
elseif Outletz<Dz || Outlety+HOutlet>=H || Outletz+DOutlet>=D
    'ERROR - Outlet does not fit within domain'
end

if Dx==Dy && Dy==Dz
else
    'ERROR - Dx/=Dy or Dy/=Dz, hence continuity equation is invalid'
end

%% Solution parameters etc.
NIterations=500000; % Max number of iterations
WSOR=1.2; % relaxation factor for SOR
BuoyancyOn=0; % (set to 1 to include buoyancy, 0 to ignore buoyancy)
OutNeumann=1; % Neumann boundary condition at outlet? 1 for yes, 0 for zero velocity gradient BC, 2 for variable velocity but fixed flow rate.
% EnvHeatLoad=0.0015; % Heat load on external walls of cold aisle (kW/m^2)
EnvHeatLoad=0;
SFSOption=1; % Server fan speed option. 1 gives fan speeds adjusting to give constant flow, dT=12.5.  2 gives constant server fan speed, designed to achieve dT=12.5K with zero external static pressure.

%% Initialise variables
Phi=zeros(nx+2,ny+2,nz+2); % velocity potential, with ghost cells at boundary
[p_c p_h p_cCount p_hCount]=deal(0);
p=p_0*ones(nx+2,ny+2,nz+2);

BCs=zeros(nx+2,ny+2,nz+2); % Boundary conditions
kcont=round(Contz/Dz+1);
T(1:nx+2,1:ny+2,1:kcont,1)=(T_0+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)/(c_p*rho_0*VIn)));
T_HA=T_0+(sum(sum(EServ_0))+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)))/(VIn*rho_0*c_p);
T(1:nx+2,1:ny+2,kcont+1:nz+2,1)=T_HA;
%% Store codes for boundary condition positions
InletPts=0;
OutletPts=0;
for j=1:ny+2
    for k=1:nz+2
        if ((j*Dy>(Inlety-0.5*HInlet) && j*Dy<(Inlety+0.5*HInlet)) && (k*Dz>(Inletz-0.5*DInlet) && k*Dz<(Inletz+0.5*DInlet))) && ((j*Dy>Outlety && j*Dy<(Outlety+HOutlet)) && (k*Dz>Outletz && k*Dz<(Outletz+DOutlet)))
            'ERROR - Inlet & Outlet overlap';
            return
        elseif ((j-1.5)*Dy>Inlety && (j-1.5)*Dy<(Inlety+HInlet)) && ((k-1.5)*Dz>Inletz && (k-1.5)*Dz<(Inletz+DInlet))
            BCs(1,j,k)=2; % 2 denotes inlet
            InletPts=InletPts+1; % This line counts the number of cells representing the inlet, which is used later to determine the required Phi values at the inlet
            BCs(nx+2,j,k)=1; % 1 denotes solid wall (opposite wall)
        elseif ((j-1.5)*Dy>Outlety && (j-1.5)*Dy<(Outlety+HOutlet)) && ((k-1.5)*Dz>Outletz && (k-1.5)*Dz<(Outletz+DOutlet))
            if OutWall==1
                BCs(1,j,k)=3; % 3 denotes outlet
                BCs(nx+2,j,k)=1; % 1 denotes solid wall (opposite wall)
            else
                BCs(nx+2,j,k)=3; % 3 denotes outlet
                BCs(1,j,k)=1; % 1 denotes solid wall (opposite wall)
            end
            OutletPts=OutletPts+1; % This line counts the number of cells representing the outlet, which is used later to determine the required Phi values at the outlet
        else
            BCs(1,j,k)=1; % 1 denotes solid wall
            BCs(nx+2,j,k)=1;
        end
        
        % rack walls and roofs
        if k>kcont && k<=kcont+RackDepth/Dz
            if ((j-1.5)*Dy>RackHeight && (j-1.5)*Dy<=RackHeight+Dy)
                for i=2:nx+1
                    BCs(i,j,k)=9; % 9=rack roof
                end
            end
        end
    end
end

for i=1:nx+2
    for k=1:nz+2
        BCs(i,1,k)=1; % 1 denotes solid wall
        BCs(i,ny+2,k)=1;
    end
end
for i=1:nx+2
    for j=1:ny+2
        BCs(i,j,1)=12; % 12 denotes centre of aisle
        BCs(i,j,nz+2)=12;
    end
end

ContFlow=0; % Resets record of flow rate through the containment partition
Vtemp=0;
for i=2:nx+1
    for j=2:ny+1
        for k=2:nz+1
            if k==kcont
                Rack=floor((((i-1.5)*Dx)+1)/0.8); % Identifies rack number
                if Rack>N_racks
                    Rack=N_racks;
                end
                if (i>=ceil(1.5+Servx(Rack)/Dx) && i<=ceil(1.5+Servx(Rack)/Dx+round(WServ(Rack)/Dx)-1)) && (j>=round(0.5*Dy+Servy(Rack)/Dy) && j<=round(0.5*Dy+Servy(Rack)/Dy)+round(HServ(Rack)/Dy)-1) % within area of a server
                    BCs(i,j,k)=5; % 5 denotes server inlet
                    ServerPts(Rack)=ServerPts(Rack)+1;
                    VServ(i,j)=VServ_0(1)/(HServ(Rack)*WServ(Rack)/(Dx*Dy)); % Initialising
                    EServ(i,j)=EServ_0(1)/(HServ(Rack)*WServ(Rack)/(Dx*Dy)); % Initialising
                    if EmptySlot==1 && (j+1-1.5)*Dy>=(Servy(Rack)+HServ(Rack))
                        BCs(i,j+1,k)=11; % denotes empty slot
                        SlotPts(Rack)=SlotPts(Rack)+1;
                    end
                    for c=1:floor(LServ(Rack)/Dz)
                        BCs(i,j,k+c)=6; % Inside server
                    end
                    BCs(i,j,k+c+1)=7; % Server outlet
                else % not within area of server
                    if BCs(i,j,k)==11 % Empty slot
                    else
                        BCs(i,j,k)=4; % 4 denotes rack front/aisle containment
                    end
                end
                
                ContFlow=ContFlow+Vtemp*Dx*Dy; % Total flow from cold to hot aisle
                Offset=0; % In Test Data Centre, rack 1 is offset 30mm from supply duct wall
                if (j-1.5)*Dy<2.1 && BCs(i,j,k)==4 % If within height of rack (2.1m) and in the plane of server inlet
                    if (max(((i-1.5)*Dx>Servx) + ((i-1.5)*Dx<Servx+WServ))==2) % within area of servers/blanking panels
                    else
                        BCs(i,j,k)=8; % leakage path
                        BypassPts=BypassPts+1;
                    end
                end
                ContFlow=ContFlow+Vtemp*Dx*Dy; % Total flow from cold to hot aisle
            end
        end
    end
end    
for i1=1:N_racks+1
    for i2=-floor(0.5*round(0.4/CellSide))+1:ceil(0.5*round(0.4/CellSide))
        for i3=2:floor((2.1/Dy)+1.5)
            if round(1+(i1-1)*0.8/CellSide+i2)>1 && round(1+(i1-1)*0.8/CellSide+i2)<nx+2
                if BCs(round(1+(i1-1)*0.8/CellSide+i2),i3,kcont)==4 % This makes sure we're not over-riding a server inlet
                    BCs(round(1+(i1-1)*0.8/CellSide+i2),i3,kcont)=8; % leakage path
                    BypassPts=BypassPts+1;
                end
            end
        end
    end
end
for i1=1:N_racks-1
    BCs(round(i1*0.8/CellSide)+1,2:round(RackHeight/CellSide+1.5)-1,kcont+1:round(kcont+RackDepth/CellSide))=10; % 10=rack side
end

istore=nx+2; % These are used in the plotting program
jstore=ny+2;
kstore=nz+2;

%% Convergence parameters etc.
iteration=2; % Iteration 1 being the initial conditions
PhiRes=1;
PhiConvergedIt=0;
PhiResTarget=0.000001;
pRes=1;
pConvergedIt=0;
pResTarget=0.004;
TRes=1;
TConvergedIt=0;
MaxTChange=5*ones(1,1);
MaxTChangeTarget=0.05;

%% Begin iteration
while iteration<=NIterations && TConvergedIt==0
%     if iteration/25==floor(iteration/25) % A monitor for how far the simulation has progessed
%         iteration
%         PhiRes
%     end
    ContFlow=0;
    ContFlowStore(iteration)=0;
    if (PhiRes>PhiResTarget || pRes>pResTarget) && PhiConvergedIt<2
        
        %% calculate bulk pressure difference between h&c aisles
        [p_c p_h p_cCount p_hCount]=deal(0);
        for i=2:nx+1 % Ignore ghost cells
            for j=2:ny+1
                ColdAisle=1;
                for k=2:nz+1
                    if k>kcont
                        ColdAisle=0;
                    end
                    if ColdAisle==1 && isnan(p(i,j,k,1))==0
                        p_c=p_c+p(i,j,k,1);
                        p_cCount=p_cCount+1;
                    elseif isnan(p(i,j,k,1))==0
                        p_h=p_h+p(i,j,k,1);
                        p_hCount=p_hCount+1;
                    end
                end
            end
        end
        dp=(p_c/p_cCount)-(p_h/p_hCount); % average dp across containment
        dpStore(iteration)=dp;
        %% Phi calculation
        % first set Phi at domain boundaries where i/j/k=1
        for j=1:ny+2 % walls in the x-plane
            for k=1:nz+2
                if BCs(1,j,k)==1 % solid wall
                    if BCs(3,j,k)==6 % If adcacent cell is adjacent to server, need to avoid Phi from this cell staying at 0 regardless of changes to Phi in adjacent cells
                        Phi(1,j,k,2)=Phi(1,j-1,k,2);
                    else
                        Phi(1,j,k,2)=Phi(2,j,k,2-1);
                    end
                elseif BCs(1,j,k)==2 % inlet
                    Phi(1,j,k,2)=Phi(2,j,k,2-1)+(VIn/(InletPts*Dy*Dz))*Dx; % Just gives constant velocity across the inlet.
                elseif BCs(1,j,k)==3 % outlet
                    if OutNeumann==1 % Neumann BC at outlet
                        Phi(1,j,k,2)=Phi(2,j,k,2-1)-(VIn/(OutletPts*Dy*Dz))*Dx; % Just gives constant velocity across the outlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(1,j,k,2)=Phi(2,j,k,2-1)+(Phi(3,j,k,2-1)-Phi(2,j,k,2-1)); % Free boundary, velocity gradient=0
                    else % Variable velocity, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(2,j:j+floor(HOutlet/Dy),k:k+floor(DOutlet/Dz),2))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(1,j,k,2)=PhiOutAv-(VIn/(OutletPts*Dy*Dz))*Dx; % Should give the right flow rate across the outlet, but allows it to fluctuate                        
                    end
                end
                if iteration==2 % only set Phi at i/j/k=nx/ny/nz, since in later iterations this calculation should use the most up to date neighbouring Phi values
                    if BCs(nx+2,j,k)==1
                        Phi(nx+2,j,k,2)=Phi(nx+1,j,k,2-1);
                    elseif BCs(nx+2,j,k)==2
                        Phi(nx+2,j,k,2)=Phi(nx+1,j,k,2-1)+(VIn/(InletPts*Dy*Dz))*Dx; % Just gives constant velocity across the inlet.
                    elseif BCs(nx+2,j,k)==3
                        Phi(nx+2,j,k,2)=Phi(nx+1,j,k,2-1)-(VIn/(OutletPts*Dy*Dz))*Dx; % Just gives constant velocity across the outlet.
                    end
                end
            end
        end
        for i=1:nx+2 % walls in the y-plane
            for k=1:nz+2
                if BCs(i,1,k)==1 % solid wall
                    if BCs(i,3,k)==6
                        Phi(i,1,k,2)=Phi(i-1,1,k,2);
                    else
                        Phi(i,1,k,2)=Phi(i,2,k,2-1);
                    end
                elseif BCs(i,1,k)==2 % inlet
                    Phi(i,1,k,2)=Phi(i,2,k,2-1)+(VIn/(InletPts*Dx*Dz))*Dy; % Just gives constant velocity across the inlet.
                elseif BCs(i,1,k)==3 % outlet
                    if OutNeumann==1 % Neumann BC at outlet
                        Phi(i,1,k,2)=Phi(i,2,k,2-1)-(VIn/(OutletPts*Dx*Dz))*Dy; % Just gives constant velocity across the outlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(i,1,k,2)=Phi(i,2,k,2-1)+(Phi(i,3,k,2-1)-Phi(i,2,k,2-1)); % Free boundary, velocity gradient=0
                    else % Variable velocity, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(i:i+floor(WOutlet/Dy),2,k:k+floor(DOutlet/Dz),2))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(i,1,k,2)=PhiOutAv-(VIn/(OutletPts*Dx*Dz))*Dy; % Should give the right flow rate across the outlet, but allows it to fluctuate
                    end
                end
                if iteration==2
                    if BCs(i,ny+2,k)==1
                        Phi(i,ny+2,k,2)=Phi(i,ny+1,k,2-1);
                    elseif BCs(i,ny+2,k)==2
                        Phi(i,ny+2,k,2)=Phi(i,ny+1,k,2-1)+(VIn/(InletPts*Dx*Dz))*Dy; % Just gives constant velocity across the inlet.
                    elseif BCs(i,ny+2,k)==3
                        Phi(i,ny+2,k,2)=Phi(i,ny+1,k,2-1)-(VIn/(OutletPts*Dx*Dz))*Dy; % Just gives constant velocity across the outlet.
                    end
                end
            end
        end
        for i=1:nx+2 % walls in the z-plane
            for j=1:ny+2
                if BCs(i,j,1)==1 % solid wall
                    if BCs(i,j,3)==6
                        Phi(i,j,1,2)=Phi(i,j-1,1,2);
                    else
                        Phi(i,j,1,2)=Phi(i,j,2,2-1);
                    end
                elseif BCs(i,j,1)==2 % inlet
                    Phi(i,j,1,2)=Phi(i,j,2,2-1)+(VIn/(InletPts*Dx*Dy))*Dz; % Just gives constant velocity across the inlet.
                elseif BCs(i,j,1)==3 % outlet
                    if OutNeumann==1 % Neumann BC at outlet
                        Phi(i,j,1,2)=Phi(i,j,2,2-1)-(VIn/(OutletPts*Dx*Dy))*Dz; % Just gives constant velocity across the inlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(i,j,1,2)=Phi(i,j,2,2-1)+(Phi(i,j,3,2-1)-Phi(i,j,2,2-1)); % Free boundary, velocity gradient=0
                    else % Variable velocity across outlet, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(i:i+floor(WOutlet/Dy),j:j+floor(HOutlet/Dy),2,2))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(i,j,1,2)=PhiOutAv-(VIn/(OutletPts*Dx*Dy))*Dz; % Should give the right flow rate across the outlet, but allows it to fluctuate
                    end
                elseif BCs(i,j,1)==12 % Centre of aisle
                    Phi(i,j,1,2)=Phi(i,j,2,1); % Zero velocity
                end
                if iteration==2
                    if BCs(i,j,nz+2)==1
                        Phi(i,j,nz+2,2)=Phi(i,j,nz+1,2-1);
                    elseif BCs(i,j,nz+2)==2
                        Phi(i,j,nz+2,2)=Phi(i,j,nz+1,2-1)+(VIn/(InletPts*Dx*Dy))*Dz; % Just gives constant velocity across the inlet.
                    elseif BCs(i,j,nz+2)==3
                        Phi(i,j,nz+2,2)=Phi(i,j,nz+1,2-1)-(VIn/(OutletPts*Dx*Dy))*Dz; % Just gives constant velocity across the outlet.
                    end
                end
            end
        end

        for i=1:nx+2
            for j=1:ny+2
                for k=2:nz+1
                    if BCs(i,j,k)==4 || BCs(i,j,k)==5 % containment
                        kstart=k;
                        break
                    end
                end
            end
        end
        for k=kstart:-1:2 % Now carry out Phi calculation for the cold aisle
            for j=2:ny+1
                for i=2:nx+1
                    if k>2
                        if k==kstart
                            Phi(i,j,k,2)=WSOR*((1/6)*(Phi(i-1,j,k,2)+Phi(i+1,j,k,2-1)+Phi(i,j-1,k,2)+Phi(i,j+1,k,2-1)+Phi(i,j,k-1,2-1)+Phi(i,j,k+1,2-1)))+(1-WSOR)*Phi(i,j,k,2-1);
                        else
                            Phi(i,j,k,2)=WSOR*((1/6)*(Phi(i-1,j,k,2)+Phi(i+1,j,k,2-1)+Phi(i,j-1,k,2)+Phi(i,j+1,k,2-1)+Phi(i,j,k-1,2-1)+Phi(i,j,k+1,2)))+(1-WSOR)*Phi(i,j,k,2-1);
                        end
                    else % Phi is available at (i,j,1) for the current iteration (i.e. at the boundary)
                        Phi(i,j,k,2)=WSOR*((1/6)*(Phi(i-1,j,k,2)+Phi(i+1,j,k,2-1)+Phi(i,j-1,k,2)+Phi(i,j+1,k,2-1)+Phi(i,j,k-1,2)+Phi(i,j,k+1,2)))+(1-WSOR)*Phi(i,j,k,2-1);
                    end
                end
            end
        end
        LeakCount=0;
        ContFlowStore(iteration)=0;
        SlotFlowStore(iteration)=0;
        ServerFanPowerTemp=0;
        for i=2:nx+1
            Rack=floor((((i-1.5)*Dx)+1)/0.8); % Identifies rack number
            for j=2:ny+1
                if BCs(i,j,kstart)==5 % if server inlet
                    VServAdd=0; % Used to store adjustment to server flow rates
                    EServAdd=0; % Used to store adjustment to server power consumptions
                    for m=1:nz+2
                        if BCs(i,j,m)==7 % server outlet
                            break
                        end
                    end
                    dptemp=p(i,j,kstart-1,1)-p(i,j,m+1,1); % Local pressure drop across server
                    dptempstoreserv(i,j)=dptemp;
                    %%%%%%% Call sub function to calculate changes to server power/flow rate resulting from dp
                    [EServAdd, VServAdd]=ServerBehaviour_thesis(ServPerRack, Fan25, Fan50, Fan75, Fan100, Rack, i, j, VServ, SFSOption, P_required, dptemp, FansPServer, VServ_0, ServerFanPowerMax, ServerPts, ServerFanPower, c_p, C_D, EServ_0);

                    % Calculate new EServ & VServ, using outputs of subfunction
                    EServ(i,j)=(EServ_0(Rack)/ServerPts(Rack))+EServAdd; % Store amendments to rack flow rates and power consumptions as calculated earlier
                    VServ(i,j)=(VServ_0(Rack)/ServerPts(Rack))+VServAdd;
                    Phi(i,j,kstart+1,PhiIt)=Phi(i,j,kstart,PhiIt)-Dz*(VServ(i,j)/(Dx*Dy));
                                        
                elseif BCs(i,j,kstart)==4 % containment, zero flow
                    Phi(i,j,kstart+1,PhiIt)=Phi(i,j,kstart,PhiIt);
                    Phi(i,j,kstart+2,PhiIt)=Phi(i,j,kstart+3,PhiIt-1);
                else % leakage path
                    if BCs(i,j,kstart)==11 % Empty slot
                        dptemp=p(i,j,kstart,1)-p(i,j,kstart+1,1); % Local pressure drop across partition
                        dptempstoreSlot(i,j)=dptemp;
                        % Calculate leakage using equations derived in
                        % Chapter 2 of Tatchell-Evans (2017, in press)
                        if dptemp<0
                            SlotFlow=((-5.6779+(5.6779^2+4*-421.52*dptemp)^0.5)/(2*-421.52))-((-35.207+(35.207^2+4*-877.71*dptemp)^0.5)/(2*-877.71));
                        else
                            SlotFlow=((-7.587+(7.587^2+4*441.88*dptemp)^0.5)/(2*441.88))-((-28.741+(28.741^2+4*1121.1*dptemp)^0.5)/(2*1121.1));
                        end
                        SlotFlowStore(iteration)=SlotFlowStore(iteration)+SlotFlow/SlotPts(Rack);
                        Phi(i,j,kstart+1,PhiIt)=Phi(i,j,kstart,PhiIt)-((SlotFlow/SlotPts(Rack))/(Dx*Dy))*Dz;
                        Phi(i,j,kstart+2,PhiIt)=Phi(i,j,kstart+3,PhiIt-1)+((SlotFlow/SlotPts(Rack))/(Dx*Dy))*Dz;
                    else % Normal leakage path
                        dptemp=p(i,j,kstart,1)-p(i,j,kstart+1,1); % Local pressure drop across partition
                        dptempstore(i,j)=dptemp;
                        % Calculate leakage using equations derived in
                        % Chapter 2 of Tatchell-Evans (2017, in press)
                        if dptemp<0
                            % Low bypass with rack interface open (Original equation gives bypass per rack, so scaled up)
                            ContFlow=-N_racks*(((-80.038+(80.038^2+4*452.89*-dptemp)^0.5)/(2*452.89))+0.75*((-56.592+(56.592^2+4*1289*-dptemp)^0.5)/(2*1289)));
                            % High bypass (original equation is for 4 racks, hence 0.25 factor added)
                            %ContFlow=-(N_racks/4)*((-3.689+(3.689^2+4*45.855*-dptemp)^0.5)/(2*45.855));
                        else
                            % Low bypass (Original equation gives bypass per rack, so scaled up)
                            ContFlow=N_racks*(((-80.038+(80.038^2+4*452.89*dptemp)^0.5)/(2*452.89))-0.75*((-56.592+(56.592^2+4*1289*dptemp)^0.5)/(2*1289)));
                            % High bypass (original equation is for 4 racks, hence 0.25 factor added)
                            %ContFlow=(N_racks/4)*((-3.689+(3.689^2+4*45.855*dptemp)^0.5)/(2*45.855));
                        end
                        ContFlowStore(iteration)=ContFlowStore(iteration)+ContFlow/BypassPts;
                        Phi(i,j,kstart+1,PhiIt)=Phi(i,j,kstart,PhiIt)-((ContFlow/BypassPts)/(Dx*Dy))*Dz;
                    end
                end
            end
        end

        ServerFlowResult=0; % Reset flow through server
        for k=kstart+2:nz+1 % Hot aisle
            for j=2:ny+1
                for i=2:nx+1
                    if BCs(i,j,k)==6 % Inside server
                        if BCs(i-1,j,k)~=6
                            Phi(i,j,k,PhiIt)=Phi(i-1,j,k,PhiIt);
                        elseif BCs(i+1,j,k)~=6
                            Phi(i,j,k,PhiIt)=Phi(i+1,j,k,PhiIt-1);
                        elseif BCs(i,j-1,k)~=6
                            Phi(i,j,k,PhiIt)=Phi(i,j-1,k,PhiIt);
                        elseif BCs(i,j+1,k)~=6
                            Phi(i,j,k,PhiIt)=Phi(i,j+1,k,PhiIt-1);
                        else
                            Phi(i,j,k,PhiIt)=NaN;
                        end
                        
                    elseif BCs(i,j,k) % Server outlet
                        Phi(i,j,k,PhiIt)=Phi(i,j,k+1,PhiIt-1)+(Phi(i,j,kstart,PhiIt)-Phi(i,j,kstart+1,PhiIt)); % Ensures flow into server=flow out of server
                    elseif BCs(i,j,k)==9 % rack roof
                        Phi(i,j,k,PhiIt)=Phi(i,j-1,k,PhiIt);
                    elseif BCs(i,j,k)==10 % rack side
                        Phi(i,j,k,PhiIt)=Phi(i-1,j,k,PhiIt);
                        
                    else % in open space i.e. no BCs apply
                        Phi(i,j,k,PhiIt)=WSOR*((1/6)*(Phi(i-1,j,k,PhiIt)+Phi(i+1,j,k,PhiIt-1)+Phi(i,j-1,k,PhiIt)+Phi(i,j+1,k,PhiIt-1)+Phi(i,j,k-1,PhiIt)+Phi(i,j,k+1,PhiIt-1)))+(1-WSOR)*Phi(i,j,k,PhiIt-1);
                    end
                end
            end
        end

        % Now domain boundaries
        PhiOutAvDone=0;
        for j=1:ny+2
            for k=1:nz+2
                if BCs(nx+2,j,k)==1
                    if BCs(nx,j,k)==6 % If adcacent cell is adjacent to server, need to avoid Phi from this cell staying at 0 regardless of changes to Phi in adjacent cells
                        Phi(nx+2,j,k,PhiIt)=Phi(nx+2,j-1,k,PhiIt);
                    else
                        Phi(nx+2,j,k,PhiIt)=Phi(nx+1,j,k,PhiIt);
                    end
                elseif BCs(nx+2,j,k)==2
                    Phi(nx+2,j,k,PhiIt)=Phi(nx+1,j,k,PhiIt)+(VIn/(InletPts*Dy*Dz))*Dx; % Just gives constant velocity across the inlet.
                elseif BCs(nx+2,j,k)==3
                    if OutNeumann==1 % Neumann BC or zero velocity gradient at outlet?
                        Phi(nx+2,j,k,PhiIt)=Phi(nx+1,j,k,PhiIt)-(VIn/(OutletPts*Dy*Dz))*Dx; % Just gives constant velocity across the inlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(nx+2,j,k,PhiIt)=Phi(nx+1,j,k,PhiIt)-(Phi(nx,j,k,PhiIt)-Phi(nx+1,j,k,PhiIt)); % Free boundary, velocity gradient=0
                    else % Variable velocity, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(nx+1,j:j+floor(HOutlet/Dy),k:k+floor(DOutlet/Dz),PhiIt))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(nx+2,j,k,PhiIt)=PhiOutAv-(VIn/(OutletPts*Dy*Dz))*Dx; % Should give the right flow rate across the outlet, but allows it to fluctuate                        
                    end
                end
            end
        end
        for i=1:nx+2
            for k=1:nz+2
                if BCs(i,ny+2,k)==1 % solid wall
                    if BCs(i,ny,k)==6
                        Phi(i,ny+2,k,PhiIt)=Phi(i-1,ny+2,k,PhiIt);
                    else
                        Phi(i,ny+2,k,PhiIt)=Phi(i,ny+1,k,PhiIt);
                    end
                elseif BCs(i,ny+2,k)==2 % inlet
                    Phi(i,ny+2,k,PhiIt)=Phi(i,ny+1,k,PhiIt)+(VIn/(InletPts*Dx*Dz))*Dy; % Just gives constant velocity across the inlet.
                elseif BCs(i,ny+2,k)==3 % outlet
                    if OutNeumann==1 % Neumann BC or zero velocity gradient at outlet
                        Phi(i,ny+2,k,PhiIt)=Phi(i,ny+1,k,PhiIt)-(VIn/(OutletPts*Dx*Dz))*Dy; % Just gives constant velocity across the outlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(i,ny+2,k,PhiIt)=Phi(i,ny+1,k,PhiIt)-(Phi(i,ny,k,PhiIt)-Phi(i,ny+1,k,PhiIt)); % Free boundary, velocity gradient=0
                    else % Variable velocity, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(i:i+floor(WOutlet/Dy),ny+1,k:k+floor(DOutlet/Dz),PhiIt))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(i,ny+2,k,PhiIt)=PhiOutAv-(VIn/(OutletPts*Dx*Dz))*Dy; % Should give the right flow rate across the outlet, but allows it to fluctuate
                    end
                end
            end
        end
        for i=1:nx+2
            for j=1:ny+2
                if BCs(i,j,nz+2)==1 % solid wall
                    if BCs(i,j,nz)==6
                        Phi(i,j,nz+2,PhiIt)=Phi(i-1,j,nz+2,PhiIt);
                    else
                        Phi(i,j,nz+2,PhiIt)=Phi(i,j,nz+1,PhiIt);
                    end
                elseif BCs(i,j,nz+2)==2 % inlet
                    Phi(i,j,nz+2,PhiIt)=Phi(i,j,nz+1,PhiIt)+(VIn/(InletPts*Dx*Dy))*Dz; % Just gives constant velocity across the inlet.
                elseif BCs(i,j,nz+2)==3 % outlet
                    if OutNeumann==1 % Neumann BC or zero velocity gradient at outlet
                        Phi(i,j,nz+2,PhiIt)=Phi(i,j,nz+1,PhiIt)-(VIn/(OutletPts*Dx*Dy))*Dz; % Just gives constant velocity across the outlet.
                    elseif OutNeumann==0 % Zero velocity gradient
                        Phi(i,j,nz+2,PhiIt)=Phi(i,j,nz+1,PhiIt)-(Phi(i,j,nz,PhiIt)-Phi(i,j,nz+1,PhiIt)); % Free boundary, velocity gradient=0
                    else % Variable velocity across outlet, fixed flow rate
                        if PhiOutAvDone==0
                            PhiOutAv=(sum(sum(sum(sum(Phi(i:i+floor(WOutlet/Dy),j:j+floor(HOutlet/Dy),nz+1,PhiIt))))))/OutletPts;
                            PhiOutAvDone=1;
                        end
                        Phi(i,j,nz+2,PhiIt)=PhiOutAv-(VIn/(OutletPts*Dx*Dy))*Dz; % Should give the right flow rate across the outlet, but allows it to fluctuate
                    end
                elseif BCs(i,j,nz+2)==12 % Centre of aisle
                    Phi(i,j,nz+2,PhiIt)=Phi(i,j,nz+1,PhiIt); % Zero velocity
                end
            end
        end

        %% Calculate velocities
[ustore vstore wstore]=deal(0);
        for i=1:nx+2 % The nodes of the velocity field correspond with the cell faces
            for j=1:ny+2
                for k=1:nz+2
                    if i<nx+2
                        u(i,j,k)=(Phi(i,j,k,PhiIt)-Phi(i+1,j,k,PhiIt))/Dx;
                        if isnan(u(i,j,k))==1
                            u(i,j,k)=0;
                        end
                    end
                    if j<ny+2
                        v(i,j,k)=(Phi(i,j,k,PhiIt)-Phi(i,j+1,k,PhiIt))/Dy;
                        if isnan(v(i,j,k))==1
                            v(i,j,k)=0;
                        end
                    end
                    if k<nz+2
                        if k==kstart+1
                            w(i,j,k)=w(i,j,k-1);
                        else
                            w(i,j,k)=(Phi(i,j,k,PhiIt)-Phi(i,j,k+1,PhiIt))/Dz;
                            if isnan(w(i,j,k))==1
                                w(i,j,k)=0;
                                wnanStore(i,j,k)=1;
                            end
                        end
                    end
                end
            end
        end

    else % if Phi field has converged, set Phi and other relevant variables at same values as previous iteration
        if PhiConvergedIt==iteration-1
            StartTCalc=[datestr(now, 'HH') datestr(now, 'MM')]
        end
        ContFlowStore(iteration)=ContFlowStore(iteration-1);
        SlotFlowStore(iteration)=SlotFlowStore(iteration-1);
        Phi(1:nx+2,1:ny+2,1:nz+2,PhiIt)=Phi(1:nx+2,1:ny+2,1:nz+2,PhiIt-1);
           
    %% advection
        maxV=(max(max(max(u))));
        if (max(max(max(v))))>maxV
            maxV=(max(max(max(v))));
        end
        if (max(max(max(w))))>maxV
            maxV=(max(max(max(w))));
        end
        T(1:nx+2,1:ny+2,1:nz+2,2)=T(1:nx+2,1:ny+2,1:nz+2,1);        
        T_HAcurrent=mean(mean(mean(mean(T(1:nx+2,1:ny+2,kstart+1+floor(RackDepth/Dz):nz+2,2))))); % calculate mean HA temperature for use in calculation of heat loss through the walls
        T_CAcurrent=mean(mean(mean(mean(T(1:nx+2,1:ny+2,1:kstart+floor(RackDepth/Dz),2))))); % calculate mean HA temperature for use in calculation of heat loss through the walls
        for i=2:nx+1 % T nodes correspond to Phi nodes, no need to calculate T in the ghost cells
            Rack=floor((((i-1.5)*Dx)+1)/0.8); % Identifies rack number
            for j=2:ny+1
                for k=2:nz+1
                    if BCs(i,j,k)==2 % Inlet to test data centre
                        T(i,j,k,2)=T_0; % maintains temperature of inlet air
                    elseif BCs(i,j,k)==6 % inside server
                        T(i,j,k,2)=T_0;
                    elseif BCs(i,j,k)==7 % Server outlet
                        if BCs(i+1,j,k)~=7 % Calculate server outlet temperatures for all outlets in this rack at this height in one go, make uniform
                            % First calculate mass-weighted mean inlet temperature
                            MeanTIn=0;
                            FlowIn=0;
                            ServerPowerTemp=0;
                            for i1=-round(WServ(Rack)/Dx)+1:0
                                MeanTIn=MeanTIn+T(i+i1,j,kstart,2)*VServ(i+i1,j);
                                FlowIn=FlowIn+VServ(i+i1,j);
                                ServerPowerTemp=ServerPowerTemp+EServ(i+i1,j);
                            end
                            MeanTIn=MeanTIn/FlowIn;
                            % Now asign outlet temperature
                            T(i-round(WServ(Rack)/Dx)+1:i,j,k,2)=MeanTIn+ServerPowerTemp/(FlowIn*rho_0*c_p);
                        end
                    else
                        Vcell=0;
                        u0=[u(i-1,j,k) -u(i,j,k)];
                        v0=[v(i,j-1,k) -v(i,j,k)];
                        w0=[w(i,j,k-1) -w(i,j,k)];
                        if isnan(u0+v0+w0)==1
                            'u0 etc nan'
                            pause(5)
                        end
                        if u(i-1,j,k)+u(i,j,k)+v(i,j-1,k)+v(i,j,k)+w(i,j,k-1)+w(i,j,k)==0 || isnan(u(i-1,j,k)+u(i,j,k)+v(i,j-1,k)+v(i,j,k)+w(i,j,k-1)+w(i,j,k))==1
                            T(i,j,k,2)=T(i,j,k,1);
                        else
                            Vcell=sum(u0)+sum(v0)+sum(w0); % This represents the excess flow into the cell, should be 0

                            HeatEnt=((u0(1)>0)*u0(1)*T(i-1,j,k,2)+(u0(2)>0)*u0(2)*T(i+1,j,k,1)+(v0(1)>0)*v0(1)*T(i,j-1,k,2)+(v0(2)>0)*v0(2)*T(i,j+1,k,1)+(w0(1)>0)*w0(1)*T(i,j,k-1,2)+(w0(2)>0)*w0(2)*T(i,j,k+1,1)); % Heat entering the cell through convection
                            HeatLeave=-((u0(1)<0)*u0(1)+(u0(2)<0)*u0(2)+(v0(1)<0)*v0(1)+(v0(2)<0)*v0(2)+(w0(1)<0)*w0(1)+(w0(2)<0)*w0(2)); % Heat leaving the cell through convection
                            T(i,j,k,2)=(HeatEnt-((k>kstart+1)*U_cube*(T_HAcurrent-T_0)/(rho_0*c_p*(Dx^2))))/HeatLeave; % applying heat loss through walls only to HA

                            i1=0;
                            Tadj=T(i,j,k,1);
                            if u(i-1,j,k)>0
                                i1=i1+1;
                                Tadj(i1)=T(i-1,j,k,2);
                            end
                            if u(i,j,k)<0
                                i1=i1+1;
                                Tadj(i1)=T(i+1,j,k,1);
                            end
                            if v(i,j-1,k)>0
                                i1=i1+1;
                                Tadj(i1)=T(i,j-1,k,2);
                            end
                            if v(i,j,k)<0
                                i1=i1+1;
                                Tadj(i1)=T(i,j+1,k,1);
                            end
                            if w(i,j,k-1)>0
                                i1=i1+1;
                                Tadj(i1)=T(i,j,k-1,2);
                            end
                            if w(i,j,k)<0
                                i1=i1+1;
                                Tadj(i1)=T(i,j,k+1,1);
                            end
                            if T(i,j,k,2)<min(Tadj)
                                Tamended(i,j,k)=min(Tadj)-T(i,j,k,2);
                                T(i,j,k,2)=min(Tadj);
                            elseif T(i,j,k,2)>max(Tadj)
                                Tamended(i,j,k)=max(Tadj)-T(i,j,k,2);
                                T(i,j,k,2)=max(Tadj);
                            end
                            if T(i,j,k,2)<T_0
%                                 'temperature below inlet'
                                Tamended(i,j,k)=T(i,j,k,2); % Stores wherever this error occurs
                                T(i,j,k,2)=T_0;
                            end
                        end
                    end
                end
            end
        end
        T(1:nx+2,1:ny+2,1,2)=T(1:nx+2,1:ny+2,2,2); % Symmetry about centre of cold aisle
        T(1:nx+2,1:ny+2,nz+2,2)=T(1:nx+2,1:ny+2,nz+1,2); % Symmetry about centre of hot aisle
        if isreal(u)==0 || isreal(v)==0 || isreal(w)==0 || isreal(p)==0
            'complex'
            pause(5)
            notcomplex=2;
        else
            notcomplex=1;
        end
 
        % Calculate T residual
        if TConvergedIt==0
            if length(MaxTChange)<iteration-1 % Initialise relevant iteration of MaxTChange if this is the first heat transfer iteration
                MaxTChange(iteration-1)=5;
            end
            if MaxTChange(iteration-1)>MaxTChangeTarget
                TRes=0;
                TSum=0;
                for i=1:nx
                    for j=1:ny
                        for k=1:nz
                            if isnan(T(i,j,k,2))==0
                                TRes=TRes+abs(T(i,j,k,2)-T(i,j,k,1));
                                TSum=TSum+abs(T(i,j,k,2));
                                TChange(i,j,k)=abs(T(i,j,k,2)-T(i,j,k,1));
                            end
                        end
                    end
                end
                MaxTChange(iteration)=max(TChange(:));
                TRes=TRes/TSum;
            else
                TConvergedIt=iteration;
            end
        end
        
        % Now add buoyancy
        if BuoyancyOn==1
            if iteration==PhiConvergedIt+1 || iteration==pConvergedIt+1
                uNB=u; % store velocity fields prior to the addition of buoyancy
                vNB=v;
                wNB=w;
                'Buoyancy kicked in'
            end
            [T_H T_H0 T_H1 T_H2 T_H3 T_H4 T_HCount T_H0Count T_H1Count T_H2Count T_H3Count T_H4Count]=deal(0);
            [v_H0 v_H1 v_H2 v_H3 v_H4 v_H0Count v_H1Count v_H2Count v_H3Count v_H4Count]=deal(0);
            H_buoy=Servy+HServ/2; % This is the height at which the Rankine Vortex is positioned (set to half the rack height in Lettieri et al., 2013).
            Buoyx=L*0.75-0.4; % Centre of vortex along length of aisle
            V_s=2; % Vortex strength (Lettieri et al. used values up to 4, noting 2 made sense since this is multiplied by H_buoy, which they had set to half the rack height, to give a characteristic length scale equal to the rack height).
            R=0.8; % This is the radius parameter for the Rankine vortex, and determines where the behaviour flips from forced vortex to potential vortex
            % Calculate average v and T in the hot aisle
            for i=2:nx+1
                Rack=floor((((i-1.5)*Dx)+1)/0.8); % Identifies rack number
                Rack=Rack-(Rack>N_racks);
                for j=2:ny+1
                    for k=((floor((Contz+LServ(Rack)+0.5*Dz)/Dz))+2):nz+1 % Hot aisle
                        if BCs(i,j,k)==7 || BCs(i,j,k)==6
                            errortest=1; % This variable will appear if the sub-section of the domain being used includes the server
                        elseif BCs(i,j,k-1)==7
                            errortest2=0;
                        end
                        if BCs(i,j,k)==0
                            T_H=T_H+T(i,j,k,2);
                            T_HCount=T_HCount+1;
                            if i<0.25*nx+1
                                T_H0=T_H0+T(i,j,k,2); % This refers to the region behind rack 1
                                T_H0Count=T_H0Count+1;
                                v_H0=v_H0+vNB(i,j-1,k-1);
                            elseif i<0.5*nx+1
                                T_H1=T_H1+T(i,j,k,2);
                                T_H1Count=T_H1Count+1;
                                v_H1=v_H1+vNB(i,j-1,k-1);
                            elseif i<0.75*nx+1
                                T_H2=T_H2+T(i,j,k,2);
                                T_H2Count=T_H2Count+1;
                                v_H2=v_H2+vNB(i,j-1,k-1);
                            else
                                T_H3=T_H3+T(i,j,k,2);
                                T_H3Count=T_H3Count+1;
                                v_H3=v_H3+vNB(i,j-1,k-1);
                            end
                        end
                    end
                end
            end
            if isreal(v_H1)==0
                'complex1'
                pause(5)
                notcomplex=2;
            else
                notcomplex=1;
            end
            T_H=T_H/T_HCount; % Average temperature in hot aisle
            T_H0=T_H0/T_H0Count; % Av temp in region behind rack 1
            T_H1=T_H1/T_H1Count;
            T_H2=T_H2/T_H2Count;
            T_H3=T_H3/T_H3Count;
            v_H0=v_H0/T_H0Count; % Average v in region behind rack 1 without buoyancy
            v_H1=v_H1/T_H1Count;
            v_H2=v_H2/T_H2Count;
            v_H3=v_H3/T_H3Count;
            v_B0=((v_H0+2*9.81*(1/T_H0)*(T_H0-T_H)*H_buoy)^2)^0.5; % From Lettieri et al, 2013.  (1/T) term represents the volumetric thermal expansion coefficient (equation 4b).
            v_B1=((v_H1+2*9.81*(1/T_H1)*(T_H1-T_H)*H_buoy)^2)^0.5;
            v_B2=((v_H2+2*9.81*(1/T_H2)*(T_H2-T_H)*H_buoy)^2)^0.5;
            v_B3=((v_H3+2*9.81*(1/T_H3)*(T_H3-T_H)*H_buoy)^2)^0.5;
            k_v0=V_s*(v_B0-v_H0); % From Lettieri et al, 2013, equation 6
            k_v1=V_s*(v_B1-v_H1);
            k_v2=V_s*(v_B2-v_H2);
            k_v3=V_s*(v_B3-v_H3);
            if isreal(v_B1)==0
                'complex2'
                pause(5)
                notcomplex=2;
            else
                notcomplex=1;
            end
            % Calculate new velocities in v and x directions
            for i=2:nx+1
                for j=2:ny+1
                    for k=((floor((Contz+RackDepth+0.5*Dz)/Dz))+2):nz+1 % Hot aisle
                        if BCs(i,j,k)==7 || BCs(i,j,k)==6
                            Errortest=1;
                        elseif BCs(i,j,k-1)==7
                            Errortest2=0;
                        end
                        if BCs(i,j,k)==0 % Since the u and v fields are aligned differently, r and its components must be calculated differently.
                            Dx3u=-((i-1)*Dx-(Buoyx)); % This is the 'x' component of 'r', the distant from the centre of the vortex to the point at which the vortex's contribution to velocity is to be calculated.
                            Dy0u=-((j-1.5)*Dy-H_buoy);
                            Dx3v=-((i-1.5)*Dx-(Buoyx));
                            Dy0v=-((j-1)*Dy-H_buoy);
                            if Dx3u>0
                                if Dy0u>0
                                    theta3u=asin(Dy0u/((Dx3u^2+Dy0u^2)^0.5));
                                else
                                    theta3u=asin(Dy0u/((Dx3u^2+Dy0u^2)^0.5));
                                end
                            else
                                if Dy0u>0
                                    theta3u=0.5*pi()+(0.5*pi()-asin(Dy0u/((Dx3u^2+Dy0u^2)^0.5)));
                                else
                                    theta3u=-pi()-asin(Dy0u/((Dx3u^2+Dy0u^2)^0.5));
                                end
                            end
                            if Dx3v>0
                                if Dy0v>0
                                    theta3v=asin(Dy0v/((Dx3v^2+Dy0v^2)^0.5));
                                else
                                    theta3v=asin(Dy0v/((Dx3v^2+Dy0v^2)^0.5));
                                end
                            else
                                if Dy0u>0
                                    theta3v=0.5*pi()+(0.5*pi()-asin(Dy0v/((Dx3v^2+Dy0v^2)^0.5)));
                                else
                                    theta3v=-pi()-asin(Dy0v/((Dx3v^2+Dy0v^2)^0.5));
                                end
                            end

                            V=0;
                            if (Dx3u^2+Dy0u^2)^0.5<R % Lettieri et al equation 5
                                V=k_v3*((Dx3u^2+Dy0u^2)^0.5)/R;
                            else
                                V=k_v3*((R/((Dx3u^2+Dy0u^2)^0.5))^2);
                            end
                                u(i,j,k)=uNB(i,j,k)+V*sin(theta3u); %
                            % Now v
                            if (Dx3v^2+Dy0v^2)^0.5<R % Lettieri et al equation 5
                                V=k_v3*((Dx3v^2+Dy0v^2)^0.5)/R;
                            else
                                V=k_v3*((R/((Dx3v^2+Dy0v^2)^0.5))^2);
                            end
                            v(i,j,k)=vNB(i,j,k)-V*cos(theta3v); %
                        end
                        if isreal(u)==0 || isreal(v)==0 || isreal(w)==0 || isreal(p)==0
                            'complex'
                            pause(5)
                            notcomplex=2;
                        else
                            notcomplex=1;
                        end
                    end
                end
            end
        end
        T(1:nx+2,1:ny+2,1:nz+2,1)=T(1:nx+2,1:ny+2,1:nz+2,2); % update T field
    end
   
    % Calculate max velocity, to allow maximum time step to be used in
    % pressure calculation
    maxV=(max(max(max(abs(u)))));
    if (max(max(max(abs(v)))))>maxV
        maxV=(max(max(max(abs(v)))));
    end
    if (max(max(max(abs(w)))))>maxV
        maxV=(max(max(max(abs(w)))));
    end
    
    if EquilCond==2 % If p_CA held constant
        if iteration<101
        elseif abs(1-(abs(sum(ContFlowStore(iteration-100:iteration-51))-sum(ContFlowStore(iteration:iteration-50)))/50)/ContFlowStore(iteration))<0.001 % If change in mean bypass from last 50 iterations to the previous 50 is less than 0.1%, i.e. if bypass flow rate is stable
            Weight=0;
            VIn=((1-Weight)*(sum(sum(VServ))+ContFlowStore(iteration)+SlotFlowStore(iteration))+Weight*VInStore(iteration-1)); % Calculate supply flow rate required for steady flow
        end
        VInStore(iteration)=VIn;
    else % VIn held constant
        if p_CA>0
            p_CA=p_CA*(H*L*Contz+(VIn-ContFlowStore(iteration)-SlotFlowStore(iteration)-sum(sum(VServ)))*(0.125*H*L*Contz/VIn))/(H*L*Contz);
        else
            p_CA=p_CA*(H*L*Contz-(VIn-ContFlowStore(iteration)-SlotFlowStore(iteration)-sum(sum(VServ)))*(0.125*H*L*Contz/VIn))/(H*L*Contz);
        end
        p_CA_store(iteration)=p_CA;        
    end
    
    for i=1:nx+2
        for j=1:ny+2
            ColdAisle=1; % Baseline pressure differs between hot and cold aisle
            for k=1:nz+2
                if k>kstart
                    ColdAisle=0;
                end
                if ColdAisle==1
                    p(i,j,k,2)=p_0+p_CA-((0.5*(u(i-(i>1),j,k)+u(i-(i==nx+2),j,k)))^2+(0.5*(v(i,j-(j>1),k)+v(i,j-(j==ny+2),k)))^2+(0.5*(w(i,j,k-(k>1))+w(i,j,k-(k==nz+2))))^2); % Apply Bernoulli's equation
                else
                    p(i,j,k,2)=p_0-((0.5*(u(i-(i>1),j,k)+u(i-(i==nx+2),j,k)))^2+(0.5*(v(i,j-(j>1),k)+v(i,j-(j==ny+2),k)))^2+(0.5*(w(i,j,k-(k>1))+w(i,j,k-(k==nz+2))))^2); % Apply Bernoulli's equation
                end
            end
        end
    end
    
    % Calculate Phi residual
    if PhiRes>PhiResTarget
        PhiRes=0;
        PhiSum=0;
        PhiChangeMax=0;
        for i=1:nx+2
            for j=1:ny+2
                for k=1:nz+2
                    if isnan(Phi(i,j,k,PhiIt))==0
                    PhiRes=PhiRes+abs(Phi(i,j,k,PhiIt)-Phi(i,j,k,PhiIt-1));
                    PhiSum=PhiSum+abs(Phi(i,j,k,PhiIt));
                    PhiChange=abs(Phi(i,j,k,PhiIt)-Phi(i,j,k,PhiIt-1));
                        if PhiChange>PhiChangeMax
                            PhiChangeMax=[i j k PhiChange];
                        end
                    end
                end
            end
        end
        if PhiSum==0
            PhiRes=1;
        else
            PhiRes=PhiRes/PhiSum;
            PhiResStore(iteration)=PhiRes;
        end
        if PhiResStore(iteration)>PhiResStore(iteration-1) && iteration>2 % If Phi residual is increasing...
            Phi(:,:,:,2)=Phi(:,:,:,1); % Reset Phi field to previous iteration
        end
        if PhiRes<=PhiResTarget
            PhiConvergedIt=iteration;
            U_cube=0;
            T(1:nx+2,1:ny+2,1:kstart,1)=(T_0+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)/(c_p*rho_0*VIn)));
            T_HA=T_0+(sum(sum(EServ))+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)))/(U_cube+VIn*rho_0*c_p);
            T(1:nx+2,1:ny+2,kstart+1:nz+2,1)=T_HA;
            U_cube=U_cube/((nx+2)*(ny+2)*(nz+2-kstart)); % Convert U_cube to heat loss per cell in the hot aisle
            U_cubeCA=U_cube*((nx+2)*(ny+2)*kstart)/((nx+2)*(ny+2)*(nz+2-kstart)); % Heat loss per cell in cold aisle
        elseif iteration>200
            if iteration/100==round(iteration/100) && mean(PhiResStore(iteration-50:iteration))>mean(PhiResStore(iteration-100:iteration-51)) % if Phi isn't converging
                if WSOR>1
                    WSOR=WSOR-0.1;
                    if WSOR<1
                        WSOR=1;
                    end
                end
            end
            if iteration/1000==round(iteration/1000)
                [iteration/1000 1000000*PhiRes WSOR]
                [datestr(now, 'HH') datestr(now, 'MM')]
                if PhiResStore(iteration)>PhiResStore(iteration-900) 
                    'Phi failed to converge'
                    PhiRes
                    PhiConvergedIt=iteration;
                    %U_cube=(6.761-((39.8-20.7)*0.134*1.2*1.005))/(39.8-20.7); % 24/1/17 test, 6.761kW heat load, average T in HA 39.8C, average T_supply=20.7C, V_supply=134l/s
                    U_cube=0;
                    T(1:nx+2,1:ny+2,1:kstart,1)=(T_0+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)/(c_p*rho_0*VIn)));
                    T_HA=T_0+(sum(sum(EServ))+(EnvHeatLoad*(L*H+H*2*Contz+Contz*L)))/(U_cube+VIn*rho_0*c_p);
                    T(1:nx+2,1:ny+2,kstart+1:nz+2,1)=T_HA;
                    U_cube=U_cube/((nx+2)*(ny+2)*(nz+2-kstart)); % Convert U_cube to heat loss per cell in the hot aisle
                    U_cubeCA=U_cube*((nx+2)*(ny+2)*kstart)/((nx+2)*(ny+2)*(nz+2-kstart)); % Heat loss per cell in cold aisle
                end
            end
        end
    end
    
    % Calculate p residual
    pRes=0;
    pSum=0;
    pChangeMax=0;
    for i=1:nx+2
        for j=1:ny+2
            for k=1:nz+2
                if isnan(p(i,j,k,2))==0
                    pRes=pRes+abs(p(i,j,k,2)-p(i,j,k,1));
                    pSum=pSum+abs(p(i,j,k,2))-p_0;
                    pChange=abs(p(i,j,k,2)-p(i,j,k,1));
                    if pChange>pChangeMax
                        pChangeMax=[i j k pChange];
                    end
                end
            end
        end
    end
    if pSum==0 || pRes==0
        pRes=1;
    else
        pRes=pRes/pSum;
        pResStore(iteration)=pRes;
    end

    if pRes<=pResTarget && pConvergedIt==0
        pConvergedIt=iteration;
    end
    p(1:nx+2,1:ny+2,1:nz+2,1)=p(1:nx+2,1:ny+2,1:nz+2,2);
    


iteration=iteration+1;
Phi(1:nx+2,1:ny+2,1:nz+2,1)=Phi(1:nx+2,1:ny+2,1:nz+2,2);
end

TimeEnd=[datestr(now, 'HH') datestr(now, 'MM')]

for i=1:N_racks
    for j=2:size(VServ,2)
        ServerFlows(i,j-1)=sum(VServ((i-1)*8+4:(i-1)*8+7,j));
    end
end

p=p-101000; % Convert pressures from absolute to gauge


% Display results
prompt='Show plots? 0 for no, 1 for yes.';
i=input(prompt);
while i==1
    prompt='1 for surface plots, 2 for arrow plots';
    j=input(prompt);
    if j==1
        uvwpplots_thesis(iteration, istore, jstore, kstore, nx, ny, nz, u, v, w, p, T); % call function for plots
    elseif j==2
        ArrowPlot_thesis(iteration, istore, jstore, kstore, nx, ny, nz, u, v, w); % call function for arrow plots
    end
    prompt='More plots? 0 for no, 1 for yes';
    i=input(prompt);
end
for i=1:nx+1
    for j=1:ny+1
        for k=1:nz+1
            V(i,j,k)=(u(i,j,k)^2+v(i,j,k)^2+w(i,j,k)^2)^0.5; % Store velocity magnitude
        end
    end
end

% Calculate flow across outlet
iteration=iteration-1;
Phi1=0;
Phi2=0;
VOut=0;
for j=1:ny+2
    for k=1:nz+2
        if BCs(nx+2,j,k)==3 % outlet
            VOut=VOut+Dy*Dz*(Phi(nx+1,j,k,PhiIt)-Phi(nx+2,j,k,PhiIt))/Dx;
        end
    end
end

% Results for comparison with experimental measurements
ServerInletTemps=0;
startSIT(3)=0;
for i=1:nx+2
    for j=1:ny+2
        for k=1:nz+2
            if BCs(i,j,k)==5
                if startSIT(1)==0
                    startSIT(1)=i;
                    startSIT(2)=j;
                    startSIT(3)=k;
                end
%                 ServerInletTemps(i-startSIT(1)+1,j-startSIT(2)+1,k-startSIT(3)+1)=T(i,j,k-1,iteration-1);
%                 ServerInletTemps(i-startSIT(1)+1,j-startSIT(2)+1,k-startSIT(3)+2)=T(i,j,k,iteration-1);
%                 ServerInletTemps(i-startSIT(1)+1,j-startSIT(2)+1,k-startSIT(3)+3)=T(i,j,k+1,iteration-1);
                ServerInletTemps(i-startSIT(1)+1,j-startSIT(2)+1)=T(i,j,k,2);
            end
        end
    end
end
csvwrite('ServerInletTemps.csv',ServerInletTemps);
for i=1:N_racks
    for j=1:size(ServerInletTemps,2)
        ServerInletTemps2(i,j)=0.25*sum(ServerInletTemps((i-1)*8+1:(i-1)*8+4,j));
    end
end
csvwrite('ServerInletTemps2.csv',ServerInletTemps2);
for i=1:N_racks
    for j=1:size(EServ,2)-1
        ServerPowers(i,j)=sum(EServ((i-1)*8+4:(i-1)*8+7,j+1));
    end
end
csvwrite('ServerPowers.csv',ServerPowers);
csvwrite('ServerFlows.csv',ServerFlows);
ServerDTs=ServerPowers./(ServerFlows*c_p*rho_0);
csvwrite('ServerDTs.csv',ServerDTs);
for i=1:N_racks
    for j=1:8
        TServersPlane(3*(i-1)+2:3*(i-1)+3,j)=ServerInletTemps2(i,j);
        TServersPlane(3*(i-1)+1,j)=NaN;
        %y(3*(i-1)+1:3*(i-1)+3)=(0.8*(i-1)+0.8/3:0.8/3:0.8*(i-1)+0.8);
        x(3*(i-1)+1:3*(i-1)+3)=[0.8*(i-1) 0.8*(i-1)+0.8*1/4 0.8*(i-1)+0.8*3/4];
        y(j)=j*0.1;
    end
end
TServersPlane(1:48,9)=TServersPlane(1:48,8);
y(9)=0.9;
surf(y,x,TServersPlane);
axis equal
view(2);
colorbar

%continuity map
for i=2:nx+1
    for j=2:ny+1
        for k=2:nz+1
                Cont(i-1,j-1,k-1)=Phi(i,j,k,PhiIt-1)-(Phi(i+1,j,k,PhiIt-1)+Phi(i-1,j,k,PhiIt-1)+Phi(i,j+1,k,PhiIt-1)+Phi(i,j-1,k,PhiIt-1)+Phi(i,j,k+1,PhiIt-1)+Phi(i,j,k-1,PhiIt-1))/6;
                ContProp(i-1,j-1,k-1)=(Phi(i,j,k,PhiIt-1)-(Phi(i+1,j,k,PhiIt-1)+Phi(i-1,j,k,PhiIt-1)+Phi(i,j+1,k,PhiIt-1)+Phi(i,j-1,k,PhiIt-1)+Phi(i,j,k+1,PhiIt-1)+Phi(i,j,k-1,PhiIt-1))/6)/(Phi(i,j,k-1,PhiIt-1)-Phi(i,j,k+1,PhiIt-1));
        end
    end
end

% Calculate bypass and recirculation
RecircMap=0;
BypassMap=0;
for i=1:nx+2
    for j=1:ny+2
        if w(i,j,kcont)<0 && BCs(i,j,kcont)~=5
            RecircMap(i,j)=w(i,j,kcont);
        elseif BCs(i,j,kcont)~=5
            BypassMap(i,j)=w(i,j,kcont);
        end
    end
end
TotalBypass=sum(sum(BypassMap))*Dx*Dy;
TotalRecirc=-sum(sum(RecircMap))*Dx*Dy;
%RecircBPMap=RecircMap+BypassMap;
ContFlow=ContFlowStore(iteration);
FlowResults=[VIn TotalBypass TotalRecirc ContFlow];
csvwrite('FlowResults.csv',FlowResults);
[Results]=SystemPF(VIn, VServ, EServ, VServ_0, EServ_0, ContFlow, ServerInletTemps, ServerInletTemps2, T_0);
csvwrite('Results.csv',Results);